/* VIDDRV.C
  This file contains low-level graphics functions which are
  platform-specific.
*/
#include "../keen.h"
#include "../keenext.h"
#include <stdarg.h>

#include <SDL.h>

char blitsurface_alloc = 0;
void VidDrv_printf(const char *str, ...);

#define WINDOW_STD_WIDTH            320
#define WINDOW_STD_HEIGHT           200
int window_width;
int window_height;
SDL_Surface *screen;                   // the actual video memory/window
SDL_Surface *ScrollSurface=NULL;       // 512x512 scroll buffer
// pointer to the surface that sb_blit is to assemble the scroll buffer into.
// if zoom=1 this is the same as "screen", else it's allocated as it's own
// buffer of 320x200.
SDL_Surface *BlitSurface=NULL;

// pointer to the line in VRAM to start blitting to when stretchblitting.
// this may not be the first line on the display as it is adjusted to
// center the image on the screen when in fullscreen.
unsigned char *VRAMPtr;

SDL_Rect dstrect;

extern FILE *log;
#define CKLOGFILENAME            "ck.log"

#define MAX_CONSOLE_MESSAGES     3
#define CONSOLE_MESSAGE_X        3
#define CONSOLE_MESSAGE_Y        3
#define CONSOLE_MESSAGE_SPACING  9
#define CONSOLE_EXPIRE_RATE      250
typedef struct stConsoleMessage
{
  char msg[80];
} stConsoleMessage;
stConsoleMessage cmsg[MAX_CONSOLE_MESSAGES];
int NumConsoleMessages = 0;
int ConsoleExpireTimer = 0;

// "Console" here refers to the capability to pop up in-game messages
// in the upper-left corner during game play ala Doom.
void DrawConsoleMessages(void)
{
int i;
int y;

 if (!NumConsoleMessages) return;
 if (!ConsoleExpireTimer)
 {
   NumConsoleMessages--;
   if (!NumConsoleMessages) return;
   ConsoleExpireTimer = CONSOLE_EXPIRE_RATE;
 }
 else ConsoleExpireTimer--;

 y = CONSOLE_MESSAGE_Y;
 for(i=0;i<NumConsoleMessages;i++)
 {
    font_draw(cmsg[i].msg, CONSOLE_MESSAGE_X, y, 1);
    y += CONSOLE_MESSAGE_SPACING;
 }
}

// removes all console messages
void DeleteConsoleMsgs(void)
{
  NumConsoleMessages = 0;
}

// adds a console msg to the top of the screen and scrolls any
// other existing messages downwards
void AddConsoleMsg(char *the_msg)
{
int i;
  for(i=MAX_CONSOLE_MESSAGES-2;i>=0;i--)
  {
    strcpy(cmsg[i+1].msg, cmsg[i].msg);
  }
  strcpy(cmsg[0].msg, the_msg);

  if (NumConsoleMessages < MAX_CONSOLE_MESSAGES) NumConsoleMessages++;
  ConsoleExpireTimer = CONSOLE_EXPIRE_RATE;
}

void sb_blit(void)
{
SDL_Rect srcrect;
char wraphoz, wrapvrt;
int save_dstx, save_dstw, save_srcx, save_srcw;
char tempbuf[80];

   dstrect.x = 0; dstrect.y = 0;
   dstrect.w = 320; dstrect.h = 200;

   srcrect.x = scrollx_buf;
   srcrect.y = scrolly_buf;
   if (scrollx_buf > (512-320))
   { // need to wrap right side
     srcrect.w = (512-scrollx_buf);
     wraphoz = 1;
   }
   else
   { // single blit for whole horizontal copy
     srcrect.w = 320;
     wraphoz = 0;
   }

   if (scrolly_buf > (512-200))
   { // need to wrap on bottom
     srcrect.h = (512-scrolly_buf);
     wrapvrt = 1;
   }
   else
   { // single blit for whole bottom copy
     srcrect.h = 200;
     wrapvrt = 0;
   }

   SDL_BlitSurface(ScrollSurface, &srcrect, BlitSurface, &dstrect);

   if (wraphoz && wrapvrt)
   {
      // first do same thing we do for wraphoz
      save_dstx = dstrect.x;
      save_dstw = dstrect.w;
      save_srcx = srcrect.x;
      save_srcw = srcrect.w;
      dstrect.x = srcrect.w;
      dstrect.w = 320 - dstrect.x;
      srcrect.x = 0;
      srcrect.w = (320 - srcrect.w);
      SDL_BlitSurface(ScrollSurface, &srcrect, BlitSurface, &dstrect);
      // now repeat for the bottom
      // (lower-right square)
      dstrect.y = srcrect.h;
      dstrect.h = 200 - dstrect.y;
      srcrect.y = 0;
      srcrect.h = (200 - srcrect.h);
      SDL_BlitSurface(ScrollSurface, &srcrect, BlitSurface, &dstrect);
      // (lower-left square)
      dstrect.x = save_dstx;
      dstrect.w = save_dstw;
      srcrect.x = save_srcx;
      srcrect.w = save_srcw;
      SDL_BlitSurface(ScrollSurface, &srcrect, BlitSurface, &dstrect);
   }
   else if (wraphoz)
   {
      dstrect.x = srcrect.w;
      dstrect.w = 320 - dstrect.x;
      srcrect.x = 0;
      srcrect.w = (320 - srcrect.w);
      SDL_BlitSurface(ScrollSurface, &srcrect, BlitSurface, &dstrect);
   }
   else if (wrapvrt)
   {
      dstrect.y = srcrect.h;
      dstrect.h = 200 - dstrect.y;
      srcrect.y = 0;
      srcrect.h = (200 - srcrect.h);
      SDL_BlitSurface(ScrollSurface, &srcrect, BlitSurface, &dstrect);
   }

   DrawConsoleMessages();
   if (showfps)
   {
     sprintf(tempbuf, "FPS: %03d", fps);
     font_draw(tempbuf, 320-3-(strlen(tempbuf)*8), 3, 1);
   }

   // if we're doing zoom then we have copied the scroll buffer into
   // another offscreen buffer, and must now stretchblit it to the screen
   if (options[OPT_ZOOM].value > 1)
   {
     scale2x(VRAMPtr, 640, BlitSurface->pixels, 320, 1, 320, 200);
   }

   SDL_Flip(screen);
}

void update_screen(void)
{
   if (options[OPT_ZOOM].value > 1)
   {
     scale2x(VRAMPtr, 640, BlitSurface->pixels, 320, 1, 320, 200);
   }
   SDL_Flip(screen);
}

// functions to directly set and retrieve pixels from the VGA display
void setpixel(int x, int y, unsigned char c)
{
 Uint8 *ubuff8;
    ubuff8 = (Uint8*) BlitSurface->pixels;
    ubuff8 += (y * BlitSurface->pitch) + x; 
    *ubuff8 = (Uint8) c;
}
unsigned char getpixel(int x, int y)
{
  return 0;
}

SDL_Color MyPalette[256];

// alter the color palette. the palette is not actually altered
// on-screen until pal_apply() is called.
void pal_set(char colour, char red, char green, char blue)
{
  MyPalette[colour].r = red;
  MyPalette[colour].g = green;
  MyPalette[colour].b = blue;
}

// applies all changes to the palette made with pal_set
void pal_apply(void)
{
  SDL_SetColors(screen, &MyPalette, 0, 256);
  SDL_SetColors(ScrollSurface, &MyPalette, 0, 256);
  if (blitsurface_alloc)
  {
    SDL_SetColors(BlitSurface, &MyPalette, 0, 256);
  }
}

// starts up the video driver
char VidDrv_Start(void)
{

  if (!scrollbuf)
  {
    VidDrv_printf("VidDrv_Start: Can not initilize SDL graphics until scroll buffer is allocated (allocmem())!\n");
    return 1;
  }
VidDrv_printf("sdl_init is getting called\n");
  /* Initialize SDL, exit if there is an error. */
  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_AUDIO) < 0)
  {
    VidDrv_printf("VidDrv_Start(): Could not initialize SDL: %s\n", SDL_GetError());
    return 1;
  }
  if (is_server)
  {
    SDL_WM_SetCaption("CloneKeen [host]", NULL);
  }
  else if (is_client)
  {
    SDL_WM_SetCaption("CloneKeen [multiplayer]", NULL);
  }
  else
  {
    SDL_WM_SetCaption("CloneKeen", NULL);
  }

  /* When the program is through executing, call SDL_Quit */
  atexit(SDL_Quit);

  return VidDrv_CreateSurfaces();
}

char VidDrv_CreateSurfaces(void)
{
unsigned int opts;
int stretch_blit_yoff;
  if (!options[OPT_ZOOM].value) options[OPT_ZOOM].value = 1;

  window_width = (WINDOW_STD_WIDTH * options[OPT_ZOOM].value);
  window_height = (WINDOW_STD_HEIGHT * options[OPT_ZOOM].value);
  stretch_blit_yoff = 0;

  /* Grab a surface on the screen */
  opts = SDL_SWSURFACE | SDL_HWPALETTE;
  if (options[OPT_FULLSCREEN].value)
  {
    opts |= SDL_FULLSCREEN;
    // center image on screen
    if (options[OPT_ZOOM].value==2)
    {
      window_width = 640;
      window_height = 480;
      stretch_blit_yoff = 40;
    }
    VidDrv_printf("fullscreen will be activated.\n");
  }
  VidDrv_printf("stretch_blit_yoff = %d\n", stretch_blit_yoff);

  VidDrv_printf("sdl_Setvideomode dimensions = %dx%d\n",window_width,window_height);
  screen = SDL_SetVideoMode(window_width, window_height, 8, opts);
  if( !screen )
  {
    VidDrv_printf("VidDrv_Start(): Couldn't create a SDL surface: %s\n", SDL_GetError());
    return 1;
  }

  VidDrv_printf("showcursor\n");
  if (options[OPT_FULLSCREEN].value)
    SDL_ShowCursor(SDL_DISABLE);
  else
    SDL_ShowCursor(SDL_ENABLE);


  VidDrv_printf("scrollsurface = creatergbsurfacefrom\n");
  ScrollSurface = SDL_CreateRGBSurfaceFrom(scrollbuf, 512, 512, 8, 512, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
  if (!ScrollSurface)
  {
     VidDrv_printf("VidDrv_Start(): Couldn't create ScrollSurface!\n");
     return 1;
  }

  if (options[OPT_ZOOM].value > 1)
  {
    VidDrv_printf("blitsurface = creatergbsurfacefrom\n");
    BlitSurface = SDL_CreateRGBSurfaceFrom(blitbuf, 320, 200, 8, 320, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);
    if (!BlitSurface)
    {
      VidDrv_printf("VidDrv_Start(): Couldn't create BlitSurface!\n");
      return 1;
    }
    blitsurface_alloc = 1;
  }
  else BlitSurface = screen;

  dstrect.x = 0; dstrect.y = 0;
  dstrect.w = WINDOW_STD_WIDTH;
  dstrect.h = WINDOW_STD_HEIGHT;

  VRAMPtr = screen->pixels + (window_width * stretch_blit_yoff);
  return 0;
}

// shuts down the video driver
void VidDrv_Stop(void)
{
  if(screen) { SDL_FreeSurface(screen); VidDrv_printf("freed screen\n"); }
  if(ScrollSurface) { SDL_FreeSurface(ScrollSurface); VidDrv_printf("freed scrollsurface\n"); }
  if(blitsurface_alloc) { blitsurface_alloc = 0; SDL_FreeSurface(BlitSurface); VidDrv_printf("freed blitsurface\n"); }
}

// resets graphics to allow changing of resolution or zoom settings
void VidDrv_reset(void)
{
     VidDrv_printf("restart: shutting down viddrv\n");
     VidDrv_Stop();

     // re-allocate memory as blitbuffer may appear/disappear
     freemem();
     if (allocmem()) { crashflag = 1; return; }

     VidDrv_printf("restart: restarting viddrv\n");
     VidDrv_CreateSurfaces();   // don't re-call sdl_init
     pal_fade(PAL_FADE_SHADES);
}

// initilize the console that VidDrv_printf outputs to. For win32
// our "console" is just a log file, so here it just clears the log.
// With futures ports this function should be expanded, as this
// file should be platform indepedent (it is supposed to be
// SDL-specific, not win32-specific)
void VidDrv_InitConsole(void)
{
FILE *fp;
  // erase contents of log file
  fp = fopen(CKLOGFILENAME, "wb");
  fclose(fp);
}

void VidDrv_printf(const char *str, ...)
{
  va_list ar;
  char buf[2048];

  va_start(ar, str);
  vsprintf(buf, str, ar);
  va_end(ar);

  #ifdef TARGET_LNX
     printf("%s", buf);
  #else
     log = fopen(CKLOGFILENAME, "a+");
     if (log)
     {
       fprintf(log, "%s", buf);
       fclose(log);
     }
  #endif
}
